@@ -255,6 +255,72 @@ setVirtualScreenSize (CompScreen *screen,
 		      int	 hsize,
 		      int	 vsize)
 {
+    /* if hsize or vsize is being reduced */
+    if (hsize < screen->hsize ||
+	vsize < screen->vsize)
+    {
+	CompWindow *w;
+	int        tx = 0;
+	int        ty = 0;
+
+	if (screen->x >= hsize)
+	    tx = screen->x - (hsize - 1);
+	if (screen->y >= vsize)
+	    ty = screen->y - (vsize - 1);
+
+	if (tx != 0 || ty != 0)
+	    moveScreenViewport (screen, tx, ty, TRUE);
+
+	/* Move windows that were in one of the deleted viewports into the
+	   closest viewport */
+	for (w = screen->windows; w; w = w->next)
+	{
+	    int moveX = 0;
+	    int moveY = 0;
+
+	    if (windowOnAllViewports (w))
+		continue;
+
+	    /* Find which viewport the (inner) window's top-left corner falls
+	       in, and check if it's outside the new viewport horizontal and
+	       vertical index range */
+	    if (hsize < screen->hsize)
+	    {
+		int vpX;   /* x index of a window's vp */
+
+		vpX = w->serverX / screen->width;
+		if (w->serverX < 0)
+		    vpX -= 1;
+
+		vpX += screen->x; /* Convert relative to absolute vp index */
+
+		/* Move windows too far right to left */
+		if (vpX >= hsize)
+		    moveX = ((hsize - 1) - vpX) * screen->width;
+	    }
+	    if (vsize < screen->vsize)
+	    {
+		int vpY;   /* y index of a window's vp */
+
+		vpY = w->serverY / screen->height;
+		if (w->serverY < 0)
+		    vpY -= 1;
+
+		vpY += screen->y; /* Convert relative to absolute vp index */
+
+		/* Move windows too far right to left */
+		if (vpY >= vsize)
+		    moveY = ((vsize - 1) - vpY) * screen->height;
+	    }
+
+	    if (moveX != 0 || moveY != 0)
+	    {
+		moveWindow (w, moveX, moveY, TRUE, TRUE);
+		syncWindowPosition (w);
+	    }
+	}
+    }
+
     screen->hsize = hsize;
     screen->vsize = vsize;
 
@@ -267,10 +333,11 @@ updateOutputDevices (CompScreen	*s)
     CompOutput	  *o, *output = NULL;
     CompListValue *list = &s->opt[COMP_SCREEN_OPTION_OUTPUTS].value.list;
     int		  nOutput = 0;
-    int		  x, y, i, bits;
+    int		  x, y, i, j, bits;
     unsigned int  width, height;
     int		  x1, y1, x2, y2;
     Region	  region;
+    CompWindow    *w;
 
     for (i = 0; i < list->nValue; i++)
     {
@@ -367,11 +434,18 @@ updateOutputDevices (CompScreen	*s)
 	free (s->outputDev);
     }
 
-    s->outputDev  = output;
-    s->nOutputDev = nOutput;
+    s->outputDev             = output;
+    s->nOutputDev            = nOutput;
+    s->hasOverlappingOutputs = FALSE;
 
     setCurrentOutput (s, s->currentOutputDev);
 
+    /* clear out fullscreen monitor hints of all windows as
+       suggested on monitor layout changes in EWMH */
+    for (w = s->windows; w; w = w->next)
+	if (w->fullscreenMonitorsSet)
+	    setWindowFullscreenMonitors (w, NULL);
+
     updateWorkareaForScreen (s);
 
     setDefaultViewport (s);
@@ -385,6 +459,17 @@ updateOutputDevices (CompScreen	*s)
 	r.rects = &r.extents;
 	r.numRects = 1;
 
+	for (i = 0; i < nOutput - 1; i++)
+	    for (j = i + 1; j < nOutput; j++)
+            {
+		XIntersectRegion (&output[i].region,
+				  &output[j].region,
+				  region);
+		if (REGION_NOT_EMPTY (region))
+		    s->hasOverlappingOutputs = TRUE;
+	    }
+	XSubtractRegion (&emptyRegion, &emptyRegion, region);
+	
 	if (s->display->nScreenInfo)
 	{
 	    for (i = 0; i < s->display->nScreenInfo; i++)
@@ -591,29 +676,10 @@ setScreenOption (CompPlugin	 *plugin,
 	    return TRUE;
 	}
 	break;
-    case COMP_SCREEN_OPTION_OPACITY_MATCHES:
-	if (compSetOptionList (o, value))
+     case COMP_SCREEN_OPTION_FORCE_INDEPENDENT:
+	if (compSetBoolOption (o, value))
 	{
-	    CompWindow *w;
-	    int	       i;
-
-	    for (i = 0; i < o->value.list.nValue; i++)
-		matchUpdate (screen->display, &o->value.list.value[i].match);
-
-	    for (w = screen->windows; w; w = w->next)
-		updateWindowOpacity (w);
-
-	    return TRUE;
-	}
-	break;
-    case COMP_SCREEN_OPTION_OPACITY_VALUES:
-	if (compSetOptionList (o, value))
-	{
-	    CompWindow *w;
-
-	    for (w = screen->windows; w; w = w->next)
-		updateWindowOpacity (w);
-
+	    updateOutputDevices (screen);
 	    return TRUE;
 	}
 	break;
@@ -632,7 +698,6 @@ const CompMetadataOptionInfo coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM] = {
     { "refresh_rate", "int", "<min>1</min>", 0, 0 },
     { "hsize", "int", "<min>1</min><max>32</max>", 0, 0 },
     { "vsize", "int", "<min>1</min><max>32</max>", 0, 0 },
-    { "opacity_step", "int", "<min>1</min>", 0, 0 },
     { "unredirect_fullscreen_windows", "bool", 0, 0, 0 },
     { "default_icon", "string", 0, 0, 0 },
     { "sync_to_vblank", "bool", 0, 0, 0 },
@@ -644,9 +709,8 @@ const CompMetadataOptionInfo coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM] = {
     { "focus_prevention_level", "int",
       RESTOSTRING (0, FOCUS_PREVENTION_LEVEL_LAST), 0, 0 },
     { "focus_prevention_match", "match", 0, 0, 0 },
-    { "opacity_matches", "list", "<type>match</type>", 0, 0 },
-    { "opacity_values", "list", "<type>int</type>", 0, 0 },
-    { "texture_compression", "bool", 0, 0, 0 }
+    { "texture_compression", "bool", 0, 0, 0 },
+    { "force_independent_output_painting", "bool", 0, 0, 0 }
 };
 
 static void
@@ -706,9 +770,10 @@ addSequence (CompScreen        *screen,
     screen->startupSequences = s;
 
     if (!screen->startupSequenceTimeoutHandle)
-	compAddTimeout (1000,
-			startupSequenceTimeout,
-			screen);
+	screen->startupSequenceTimeoutHandle =
+	    compAddTimeout (1000, 1500,
+			    startupSequenceTimeout,
+			    screen);
 
     updateStartupFeedback (screen);
 }
@@ -748,6 +813,28 @@ removeSequence (CompScreen        *screen,
     updateStartupFeedback (screen);
 }
 
+static void
+removeAllSequences (CompScreen *screen)
+{
+    CompStartupSequence *s;
+    CompStartupSequence *sNext;
+
+    for (s = screen->startupSequences; s; s = sNext)
+    {
+	sNext = s->next;
+	sn_startup_sequence_unref (s->sequence);
+	free (s);
+    }
+    screen->startupSequences = NULL;
+
+    if (screen->startupSequenceTimeoutHandle)
+    {
+	compRemoveTimeout (screen->startupSequenceTimeoutHandle);
+	screen->startupSequenceTimeoutHandle = 0;
+    }
+    updateStartupFeedback (screen);
+}
+
 static void
 compScreenSnEvent (SnMonitorEvent *event,
 		   void           *userData)
@@ -914,7 +1001,7 @@ reshape (CompScreen *s,
     s->fullscreenOutput.workArea.y       = 0;
     s->fullscreenOutput.workArea.width   = w;
     s->fullscreenOutput.workArea.height  = h;
- 
+
     updateScreenEdges (s);
 }
 
@@ -986,7 +1073,7 @@ updateScreenBackground (CompScreen  *screen,
 				     &actualType, &actualFormat, &nItems,
 				     &bytesAfter, &prop);
 
-	if (status == Success && nItems && prop)
+	if (status == Success && prop)
 	{
 	    if (actualType   == pixmapAtom &&
 		actualFormat == 32         &&
@@ -1026,7 +1113,7 @@ updateScreenBackground (CompScreen  *screen,
 	if (!bindPixmapToTexture (screen, texture, pixmap,
 				  width, height, depth))
 	{
-	    compLogMessage (NULL, "core", CompLogLevelWarn,
+	    compLogMessage ("core", CompLogLevelWarn,
 			    "Couldn't bind background pixmap 0x%x to "
 			    "texture", (int) pixmap);
 	}
@@ -1112,104 +1199,114 @@ setSupportingWmCheck (CompScreen *s)
 		     (unsigned char *) &s->grabWindow, 1);
 }
 
-static void
-setSupported (CompScreen *s)
+static unsigned int
+addSupportedAtoms (CompScreen   *s,
+		   Atom         *atoms,
+		   unsigned int size)
 {
-    CompDisplay *d = s->display;
-    Atom	data[256];
-    int		i = 0;
+    CompDisplay  *d = s->display;
+    unsigned int count = 0;
 
-    data[i++] = d->supportedAtom;
-    data[i++] = d->supportingWmCheckAtom;
+    atoms[count++] = d->utf8StringAtom;
 
-    data[i++] = d->utf8StringAtom;
+    atoms[count++] = d->clientListAtom;
+    atoms[count++] = d->clientListStackingAtom;
 
-    data[i++] = d->clientListAtom;
-    data[i++] = d->clientListStackingAtom;
+    atoms[count++] = d->winActiveAtom;
 
-    data[i++] = d->winActiveAtom;
+    atoms[count++] = d->desktopViewportAtom;
+    atoms[count++] = d->desktopGeometryAtom;
+    atoms[count++] = d->currentDesktopAtom;
+    atoms[count++] = d->numberOfDesktopsAtom;
+    atoms[count++] = d->showingDesktopAtom;
 
-    data[i++] = d->desktopViewportAtom;
-    data[i++] = d->desktopGeometryAtom;
-    data[i++] = d->currentDesktopAtom;
-    data[i++] = d->numberOfDesktopsAtom;
-    data[i++] = d->showingDesktopAtom;
+    atoms[count++] = d->workareaAtom;
 
-    data[i++] = d->workareaAtom;
+    atoms[count++] = d->wmNameAtom;
 
-    data[i++] = d->wmNameAtom;
-/*
-    data[i++] = d->wmVisibleNameAtom;
-*/
+    atoms[count++] = d->wmStrutAtom;
+    atoms[count++] = d->wmStrutPartialAtom;
 
-    data[i++] = d->wmStrutAtom;
-    data[i++] = d->wmStrutPartialAtom;
+    atoms[count++] = d->wmUserTimeAtom;
+    atoms[count++] = d->frameExtentsAtom;
+    atoms[count++] = d->frameWindowAtom;
 
-/*
-    data[i++] = d->wmPidAtom;
-*/
+    atoms[count++] = d->winStateAtom;
+    atoms[count++] = d->winStateModalAtom;
+    atoms[count++] = d->winStateStickyAtom;
+    atoms[count++] = d->winStateMaximizedVertAtom;
+    atoms[count++] = d->winStateMaximizedHorzAtom;
+    atoms[count++] = d->winStateShadedAtom;
+    atoms[count++] = d->winStateSkipTaskbarAtom;
+    atoms[count++] = d->winStateSkipPagerAtom;
+    atoms[count++] = d->winStateHiddenAtom;
+    atoms[count++] = d->winStateFullscreenAtom;
+    atoms[count++] = d->winStateAboveAtom;
+    atoms[count++] = d->winStateBelowAtom;
+    atoms[count++] = d->winStateDemandsAttentionAtom;
 
-    data[i++] = d->wmUserTimeAtom;
-    data[i++] = d->frameExtentsAtom;
-    data[i++] = d->frameWindowAtom;
-
-    data[i++] = d->winStateAtom;
-    data[i++] = d->winStateModalAtom;
-    data[i++] = d->winStateStickyAtom;
-    data[i++] = d->winStateMaximizedVertAtom;
-    data[i++] = d->winStateMaximizedHorzAtom;
-    data[i++] = d->winStateShadedAtom;
-    data[i++] = d->winStateSkipTaskbarAtom;
-    data[i++] = d->winStateSkipPagerAtom;
-    data[i++] = d->winStateHiddenAtom;
-    data[i++] = d->winStateFullscreenAtom;
-    data[i++] = d->winStateAboveAtom;
-    data[i++] = d->winStateBelowAtom;
-    data[i++] = d->winStateDemandsAttentionAtom;
-
-    data[i++] = d->winOpacityAtom;
-    data[i++] = d->winBrightnessAtom;
+    atoms[count++] = d->winOpacityAtom;
+    atoms[count++] = d->winBrightnessAtom;
 
     if (s->canDoSaturated)
     {
-	data[i++] = d->winSaturationAtom;
-	data[i++] = d->winStateDisplayModalAtom;
+	atoms[count++] = d->winSaturationAtom;
+	atoms[count++] = d->winStateDisplayModalAtom;
     }
 
-    data[i++] = d->wmAllowedActionsAtom;
+    atoms[count++] = d->wmAllowedActionsAtom;
 
-    data[i++] = d->winActionMoveAtom;
-    data[i++] = d->winActionResizeAtom;
-    data[i++] = d->winActionStickAtom;
-    data[i++] = d->winActionMinimizeAtom;
-    data[i++] = d->winActionMaximizeHorzAtom;
-    data[i++] = d->winActionMaximizeVertAtom;
-    data[i++] = d->winActionFullscreenAtom;
-    data[i++] = d->winActionCloseAtom;
-    data[i++] = d->winActionShadeAtom;
-    data[i++] = d->winActionChangeDesktopAtom;
-    data[i++] = d->winActionAboveAtom;
-    data[i++] = d->winActionBelowAtom;
+    atoms[count++] = d->winActionMoveAtom;
+    atoms[count++] = d->winActionResizeAtom;
+    atoms[count++] = d->winActionStickAtom;
+    atoms[count++] = d->winActionMinimizeAtom;
+    atoms[count++] = d->winActionMaximizeHorzAtom;
+    atoms[count++] = d->winActionMaximizeVertAtom;
+    atoms[count++] = d->winActionFullscreenAtom;
+    atoms[count++] = d->winActionCloseAtom;
+    atoms[count++] = d->winActionShadeAtom;
+    atoms[count++] = d->winActionChangeDesktopAtom;
+    atoms[count++] = d->winActionAboveAtom;
+    atoms[count++] = d->winActionBelowAtom;
 
-    data[i++] = d->winTypeAtom;
-    data[i++] = d->winTypeDesktopAtom;
-    data[i++] = d->winTypeDockAtom;
-    data[i++] = d->winTypeToolbarAtom;
-    data[i++] = d->winTypeMenuAtom;
-    data[i++] = d->winTypeSplashAtom;
-    data[i++] = d->winTypeDialogAtom;
-    data[i++] = d->winTypeUtilAtom;
-    data[i++] = d->winTypeNormalAtom;
+    atoms[count++] = d->winTypeAtom;
+    atoms[count++] = d->winTypeDesktopAtom;
+    atoms[count++] = d->winTypeDockAtom;
+    atoms[count++] = d->winTypeToolbarAtom;
+    atoms[count++] = d->winTypeMenuAtom;
+    atoms[count++] = d->winTypeSplashAtom;
+    atoms[count++] = d->winTypeDialogAtom;
+    atoms[count++] = d->winTypeUtilAtom;
+    atoms[count++] = d->winTypeNormalAtom;
 
-    data[i++] = d->wmDeleteWindowAtom;
-    data[i++] = d->wmPingAtom;
+    atoms[count++] = d->wmDeleteWindowAtom;
+    atoms[count++] = d->wmPingAtom;
 
-    data[i++] = d->wmMoveResizeAtom;
-    data[i++] = d->moveResizeWindowAtom;
-    data[i++] = d->restackWindowAtom;
+    atoms[count++] = d->wmMoveResizeAtom;
+    atoms[count++] = d->moveResizeWindowAtom;
+    atoms[count++] = d->restackWindowAtom;
+
+    atoms[count++] = d->wmFullscreenMonitorsAtom;
+
+    assert (count < size);
+
+    return count;
+}
+
+void
+setSupportedWmHints (CompScreen *s)
+{
+    CompDisplay  *d = s->display;
+    Atom	 data[256];
+    unsigned int count = 0;
+
+    data[count++] = d->supportedAtom;
+    data[count++] = d->supportingWmCheckAtom;
+
+    count += (*s->addSupportedAtoms) (s, data + count, 256 - count);
 
     XChangeProperty (d->display, s->root, d->supportedAtom, XA_ATOM, 32,
-		     PropModeReplace, (unsigned char *) data, i);
+		     PropModeReplace, (unsigned char *) data, count);
 }
 
 static void
@@ -1222,18 +1319,42 @@ getDesktopHints (CompScreen *s)
     unsigned long n, left;
     unsigned char *propData;
 
-    result = XGetWindowProperty (s->display->display, s->root,
-				 d->numberOfDesktopsAtom, 0L, 1L, FALSE,
-				 XA_CARDINAL, &actual, &format,
-				 &n, &left, &propData);
-
-    if (result == Success && n && propData && useDesktopHints)
+    if (useDesktopHints)
     {
-	memcpy (data, propData, sizeof (unsigned long));
-	XFree (propData);
+	result = XGetWindowProperty (s->display->display, s->root,
+				     d->numberOfDesktopsAtom, 0L, 1L, FALSE,
+				     XA_CARDINAL, &actual, &format,
+				     &n, &left, &propData);
 
-	if (data[0] > 0 && data[0] < 0xffffffff)
-	    s->nDesktop = data[0];
+	if (result == Success && propData)
+	{
+	    if (n)
+	    {
+		memcpy (data, propData, sizeof (unsigned long));
+
+		if (data[0] > 0 && data[0] < 0xffffffff)
+		    s->nDesktop = data[0];
+	    }
+	    XFree (propData);
+	}
+
+	result = XGetWindowProperty (s->display->display, s->root,
+				     d->currentDesktopAtom, 0L, 1L, FALSE,
+				     XA_CARDINAL, &actual, &format,
+				     &n, &left, &propData);
+
+	if (result == Success && propData)
+	{
+	    if (n)
+	    {
+		memcpy (data, propData, sizeof (unsigned long));
+
+		if (data[0] < s->nDesktop)
+		    s->currentDesktop = data[0];
+	    }
+
+	    XFree (propData);
+	}
     }
 
     result = XGetWindowProperty (s->display->display, s->root,
@@ -1241,7 +1362,7 @@ getDesktopHints (CompScreen *s)
 				 FALSE, XA_CARDINAL, &actual, &format,
 				 &n, &left, &propData);
 
-    if (result == Success && n && propData)
+    if (result == Success && propData)
     {
 	if (n == 2)
 	{
@@ -1257,32 +1378,21 @@ getDesktopHints (CompScreen *s)
 	XFree (propData);
     }
 
-    result = XGetWindowProperty (s->display->display, s->root,
-				 d->currentDesktopAtom, 0L, 1L, FALSE,
-				 XA_CARDINAL, &actual, &format,
-				 &n, &left, &propData);
-
-    if (result == Success && n && propData && useDesktopHints)
-    {
-	memcpy (data, propData, sizeof (unsigned long));
-	XFree (propData);
-
-	if (data[0] < s->nDesktop)
-	    s->currentDesktop = data[0];
-    }
-
     result = XGetWindowProperty (s->display->display, s->root,
 				 d->showingDesktopAtom, 0L, 1L, FALSE,
 				 XA_CARDINAL, &actual, &format,
 				 &n, &left, &propData);
 
-    if (result == Success && n && propData)
+    if (result == Success && propData)
     {
-	memcpy (data, propData, sizeof (unsigned long));
-	XFree (propData);
+	if (n)
+	{
+	    memcpy (data, propData, sizeof (unsigned long));
 
-	if (data[0])
-	    (*s->enterShowDesktopMode) (s);
+	    if (data[0])
+		(*s->enterShowDesktopMode) (s);
+	}
+	XFree (propData);
     }
 
     data[0] = s->currentDesktop;
@@ -1431,7 +1541,8 @@ enterShowDesktopMode (CompScreen *s)
 	if ((s->showingDesktopMask & w->wmType) &&
 	    (!(w->state & CompWindowStateSkipTaskbarMask) || st->value.b))
 	{
-	    if (!w->inShowDesktopMode && !w->grabbed && (*s->focusWindow) (w))
+	    if (!w->inShowDesktopMode && !w->grabbed &&
+		w->managed && (*s->focusWindow) (w))
 	    {
 		w->inShowDesktopMode = TRUE;
 		hideWindow (w);
@@ -1538,10 +1649,10 @@ initWindowWalker (CompScreen *screen,
 static void
 freeScreen (CompScreen *s)
 {
+    int i, j;
+
     if (s->outputDev)
     {
-	int i;
-
 	for (i = 0; i < s->nOutputDev; i++)
 	    if (s->outputDev[i].name)
 		free (s->outputDev[i].name);
@@ -1567,6 +1678,19 @@ freeScreen (CompScreen *s)
     if (s->damage)
 	XDestroyRegion (s->damage);
 
+    if (s->grabs)
+	free (s->grabs);
+
+    if (s->exposeRects)
+	free (s->exposeRects);
+
+    /* XXX: Maybe we should free all fragment functions here? But
+       the definition of CompFunction is private to fragment.c ... */
+    for (i = 0; i < 2; i++)
+	for (j = 0; j < 64; j++)
+	    if (s->saturateFunction[i][j])
+		destroyFragmentFunction (s, s->saturateFunction[i][j]);
+
     compFiniScreenOptions (s, s->opt, COMP_SCREEN_OPTION_NUM);
 
     if (s->windowPrivateIndices)
@@ -1749,7 +1873,6 @@ addScreen (CompDisplay *display,
     s->enableOutputClipping	   = enableOutputClipping;
     s->disableOutputClipping	   = disableOutputClipping;
     s->applyScreenTransform	   = applyScreenTransform;
-    s->paintBackground		   = paintBackground;
     s->paintWindow		   = paintWindow;
     s->drawWindow		   = drawWindow;
     s->addWindowGeometry	   = addWindowGeometry;
@@ -1776,6 +1899,7 @@ addScreen (CompDisplay *display,
     s->windowStateChangeNotify = windowStateChangeNotify;
 
     s->outputChangeNotify = outputChangeNotify;
+    s->addSupportedAtoms  = addSupportedAtoms;
 
     s->initWindowWalker = initWindowWalker;
 
@@ -1798,7 +1922,7 @@ addScreen (CompDisplay *display,
     visinfo = XGetVisualInfo (dpy, VisualIDMask, &templ, &nvisinfo);
     if (!nvisinfo)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"Couldn't get visual info for default visual");
 	return FALSE;
     }
@@ -1809,8 +1933,7 @@ addScreen (CompDisplay *display,
 
     if (!XAllocColor (dpy, s->colormap, &black))
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
-			"Couldn't allocate color");
+	compLogMessage ("core", CompLogLevelFatal, "Couldn't allocate color");
 	XFree (visinfo);
 	return FALSE;
     }
@@ -1818,8 +1941,7 @@ addScreen (CompDisplay *display,
     bitmap = XCreateBitmapFromData (dpy, s->root, &data, 1, 1);
     if (!bitmap)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
-			"Couldn't create bitmap");
+	compLogMessage ("core", CompLogLevelFatal, "Couldn't create bitmap");
 	XFree (visinfo);
 	return FALSE;
     }
@@ -1828,7 +1950,7 @@ addScreen (CompDisplay *display,
 					      &black, &black, 0, 0);
     if (!s->invisibleCursor)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"Couldn't create invisible cursor");
 	XFree (visinfo);
 	return FALSE;
@@ -1840,7 +1962,7 @@ addScreen (CompDisplay *display,
     glXGetConfig (dpy, visinfo, GLX_USE_GL, &value);
     if (!value)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"Root visual is not a GL visual");
 	XFree (visinfo);
 	return FALSE;
@@ -1849,7 +1971,7 @@ addScreen (CompDisplay *display,
     glXGetConfig (dpy, visinfo, GLX_DOUBLEBUFFER, &value);
     if (!value)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"Root visual is not a double buffered GL visual");
 	XFree (visinfo);
 	return FALSE;
@@ -1858,8 +1980,7 @@ addScreen (CompDisplay *display,
     s->ctx = glXCreateContext (dpy, visinfo, NULL, !indirectRendering);
     if (!s->ctx)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
-			"glXCreateContext failed");
+	compLogMessage ("core", CompLogLevelFatal, "glXCreateContext failed");
 	XFree (visinfo);
 
 	return FALSE;
@@ -1868,7 +1989,7 @@ addScreen (CompDisplay *display,
     glxExtensions = glXQueryExtensionsString (dpy, screenNum);
     if (!strstr (glxExtensions, "GLX_EXT_texture_from_pixmap"))
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"GLX_EXT_texture_from_pixmap is missing");
 	XFree (visinfo);
 
@@ -1879,7 +2000,7 @@ addScreen (CompDisplay *display,
 
     if (!strstr (glxExtensions, "GLX_SGIX_fbconfig"))
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"GLX_SGIX_fbconfig is missing");
 	return FALSE;
     }
@@ -1898,17 +2019,19 @@ addScreen (CompDisplay *display,
 	getProcAddress (s, "glXGetFBConfigAttrib");
     s->createPixmap = (GLXCreatePixmapProc)
 	getProcAddress (s, "glXCreatePixmap");
+    s->destroyPixmap = (GLXDestroyPixmapProc)
+	getProcAddress (s, "glXDestroyPixmap");
 
     if (!s->bindTexImage)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"glXBindTexImageEXT is missing");
 	return FALSE;
     }
 
     if (!s->releaseTexImage)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"glXReleaseTexImageEXT is missing");
 	return FALSE;
     }
@@ -1916,9 +2039,10 @@ addScreen (CompDisplay *display,
     if (!s->queryDrawable     ||
 	!s->getFBConfigs      ||
 	!s->getFBConfigAttrib ||
-	!s->createPixmap)
+	!s->createPixmap      ||
+	!s->destroyPixmap)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"fbconfig functions missing");
 	return FALSE;
     }
@@ -1945,7 +2069,7 @@ addScreen (CompDisplay *display,
     glExtensions = (const char *) glGetString (GL_EXTENSIONS);
     if (!glExtensions)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"No valid GL extensions string found.");
 	return FALSE;
     }
@@ -1975,7 +2099,7 @@ addScreen (CompDisplay *display,
 
     if (!(s->textureRectangle || s->textureNonPowerOfTwo))
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"Support for non power of two textures missing");
 	return FALSE;
     }
@@ -1997,6 +2121,10 @@ addScreen (CompDisplay *display,
 	strstr (glExtensions, "GL_SGIS_texture_border_clamp"))
 	s->textureBorderClamp = 1;
 
+    s->activeTexture       = NULL;
+    s->clientActiveTexture = NULL;
+    s->multiTexCoord2f     = NULL;
+
     s->maxTextureUnits = 1;
     if (strstr (glExtensions, "GL_ARB_multitexture"))
     {
@@ -2011,6 +2139,14 @@ addScreen (CompDisplay *display,
 	    glGetIntegerv (GL_MAX_TEXTURE_UNITS_ARB, &s->maxTextureUnits);
     }
 
+    s->genPrograms             = NULL;
+    s->deletePrograms          = NULL;
+    s->bindProgram             = NULL;
+    s->programString           = NULL;
+    s->programEnvParameter4f   = NULL;
+    s->programLocalParameter4f = NULL;
+    s->getProgramiv            = NULL;
+
     s->fragmentProgram = 0;
     if (strstr (glExtensions, "GL_ARB_fragment_program"))
     {
@@ -2026,16 +2162,26 @@ addScreen (CompDisplay *display,
 	    getProcAddress (s, "glProgramEnvParameter4fARB");
 	s->programLocalParameter4f = (GLProgramParameter4fProc)
 	    getProcAddress (s, "glProgramLocalParameter4fARB");
+	s->getProgramiv = (GLGetProgramivProc)
+	    getProcAddress (s, "glGetProgramivARB");
 
-	if (s->genPrograms	     &&
-	    s->deletePrograms	     &&
-	    s->bindProgram	     &&
-	    s->programString	     &&
-	    s->programEnvParameter4f &&
-	    s->programLocalParameter4f)
+	if (s->genPrograms	       &&
+	    s->deletePrograms	       &&
+	    s->bindProgram	       &&
+	    s->programString	       &&
+	    s->programEnvParameter4f   &&
+	    s->programLocalParameter4f &&
+	    s->getProgramiv)
 	    s->fragmentProgram = 1;
     }
 
+    s->genFramebuffers        = NULL;
+    s->deleteFramebuffers     = NULL;
+    s->bindFramebuffer        = NULL;
+    s->checkFramebufferStatus = NULL;
+    s->framebufferTexture2D   = NULL;
+    s->generateMipmap         = NULL;
+
     s->fbo = 0;
     if (strstr (glExtensions, "GL_EXT_framebuffer_object"))
     {
@@ -2208,7 +2354,7 @@ addScreen (CompDisplay *display,
 
     if (!s->glxPixmapFBConfigs[defaultDepth].fbConfig)
     {
-	compLogMessage (display, "core", CompLogLevelFatal,
+	compLogMessage ("core", CompLogLevelFatal,
 			"No GLXFBConfig for default depth, "
 			"this isn't going to work.");
 	return FALSE;
@@ -2287,6 +2433,11 @@ addScreen (CompDisplay *display,
 	}
     }
 
+    /* enforce restack on all windows */
+    for (i = 0, w = s->reverseWindows; w && i < nchildren; i++, w = w->prev)
+	children[i] = w->id;
+    XRestackWindows (dpy, children, i);
+
     XFree (children);
 
     attrib.override_redirect = 1;
@@ -2323,7 +2474,7 @@ addScreen (CompDisplay *display,
 
     setDesktopHints (s);
     setSupportingWmCheck (s);
-    setSupported (s);
+    setSupportedWmHints (s);
 
     s->normalCursor = XCreateFontCursor (dpy, XC_left_ptr);
     s->busyCursor   = XCreateFontCursor (dpy, XC_watch);
@@ -2353,6 +2504,8 @@ removeScreen (CompScreen *s)
     else
 	d->screens = NULL;
 
+    removeAllSequences (s);
+
     while (s->windows)
 	removeWindow (s->windows);
 
@@ -2397,6 +2550,13 @@ damageScreenRegion (CompScreen *screen,
     XUnionRegion (screen->damage, region, screen->damage);
 
     screen->damageMask |= COMP_SCREEN_DAMAGE_REGION_MASK;
+
+    /* if the number of damage rectangles grows two much between repaints,
+       we have a lot of overhead just for doing the damage tracking -
+       in order to make sure we're not having too much overhead, damage
+       the whole screen if we have a lot of damage rects */
+    if (screen->damage->numRects > 100)
+	damageScreen (screen);
 }
 
 void
@@ -2433,12 +2593,39 @@ focusDefaultWindow (CompScreen *s)
     if (!d->opt[COMP_DISPLAY_OPTION_CLICK_TO_FOCUS].value.b)
     {
 	w = findTopLevelWindowAtDisplay (d, d->below);
-	if (w && !(w->type & (CompWindowTypeDesktopMask |
-			      CompWindowTypeDockMask)))
+
+	if (w && (*w->screen->focusWindow) (w))
 	{
-	    if ((*w->screen->focusWindow) (w))
+	    if (!(w->type & (CompWindowTypeDesktopMask |
+			    CompWindowTypeDockMask)))
 		focus = w;
 	}
+	else
+	{
+	    Bool         status;
+	    Window       rootReturn, childReturn;
+	    int          dummyInt;
+	    unsigned int dummyUInt;
+
+	    /* huh, we didn't find d->below ... perhaps it's out of date;
+	       try grabbing it through the server */
+
+	    status = XQueryPointer (d->display, s->root, &rootReturn,
+				    &childReturn, &dummyInt, &dummyInt,
+				    &dummyInt, &dummyInt, &dummyUInt);
+
+	    if (status && rootReturn == s->root)
+	    {
+		w = findTopLevelWindowAtDisplay (d, childReturn);
+
+		if (w && (*w->screen->focusWindow) (w))
+		{
+		    if (!(w->type & (CompWindowTypeDesktopMask |
+				     CompWindowTypeDockMask)))
+			focus = w;
+		}
+	    }
+	}
     }
 
     if (!focus)
@@ -3094,66 +3281,73 @@ computeWorkareaForBox (CompScreen *s,
 	if (!w->mapNum)
 	    continue;
 
-	if (w->struts)
+	if (!w->struts)
+	    continue;
+
+	r.extents.y1 = pBox->y1;
+	r.extents.y2 = pBox->y2;
+
+	x1 = w->struts->left.x;
+	y1 = w->struts->left.y;
+	x2 = x1 + w->struts->left.width;
+	y2 = y1 + w->struts->left.height;
+
+	if (y1 < pBox->y2 && y2 > pBox->y1)
 	{
-	    r.extents.y1 = pBox->y1;
-	    r.extents.y2 = pBox->y2;
+	    r.extents.x1 = x1;
+	    r.extents.x2 = x2;
 
-	    x1 = w->struts->left.x;
-	    y1 = w->struts->left.y;
-	    x2 = x1 + w->struts->left.width;
-	    y2 = y1 + w->struts->left.height;
-
-	    if (y1 < pBox->y2 && y2 > pBox->y1)
-	    {
-		r.extents.x1 = x1;
-		r.extents.x2 = x2;
-
-		XSubtractRegion (region, &r, region);
-	    }
-
-	    x1 = w->struts->right.x;
-	    y1 = w->struts->right.y;
-	    x2 = x1 + w->struts->right.width;
-	    y2 = y1 + w->struts->right.height;
-
-	    if (y1 < pBox->y2 && y2 > pBox->y1)
-	    {
-		r.extents.x1 = x1;
-		r.extents.x2 = x2;
-
-		XSubtractRegion (region, &r, region);
-	    }
-
-	    r.extents.x1 = pBox->x1;
-	    r.extents.x2 = pBox->x2;
-
-	    x1 = w->struts->top.x;
-	    y1 = w->struts->top.y;
-	    x2 = x1 + w->struts->top.width;
-	    y2 = y1 + w->struts->top.height;
-
-	    if (x1 < pBox->x2 && x2 > pBox->x1)
-	    {
-		r.extents.y1 = y1;
-		r.extents.y2 = y2;
-
-		XSubtractRegion (region, &r, region);
-	    }
-
-	    x1 = w->struts->bottom.x;
-	    y1 = w->struts->bottom.y;
-	    x2 = x1 + w->struts->bottom.width;
-	    y2 = y1 + w->struts->bottom.height;
-
-	    if (x1 < pBox->x2 && x2 > pBox->x1)
-	    {
-		r.extents.y1 = y1;
-		r.extents.y2 = y2;
-
-		XSubtractRegion (region, &r, region);
-	    }
+	    XSubtractRegion (region, &r, region);
 	}
+
+	x1 = w->struts->right.x;
+	y1 = w->struts->right.y;
+	x2 = x1 + w->struts->right.width;
+	y2 = y1 + w->struts->right.height;
+
+	if (y1 < pBox->y2 && y2 > pBox->y1)
+	{
+	    r.extents.x1 = x1;
+	    r.extents.x2 = x2;
+
+	    XSubtractRegion (region, &r, region);
+	}
+
+	r.extents.x1 = pBox->x1;
+	r.extents.x2 = pBox->x2;
+
+	x1 = w->struts->top.x;
+	y1 = w->struts->top.y;
+	x2 = x1 + w->struts->top.width;
+	y2 = y1 + w->struts->top.height;
+
+	if (x1 < pBox->x2 && x2 > pBox->x1)
+	{
+	    r.extents.y1 = y1;
+	    r.extents.y2 = y2;
+
+	    XSubtractRegion (region, &r, region);
+	}
+
+	x1 = w->struts->bottom.x;
+	y1 = w->struts->bottom.y;
+	x2 = x1 + w->struts->bottom.width;
+	y2 = y1 + w->struts->bottom.height;
+
+	if (x1 < pBox->x2 && x2 > pBox->x1)
+	{
+	    r.extents.y1 = y1;
+	    r.extents.y2 = y2;
+
+	    XSubtractRegion (region, &r, region);
+	}
+    }
+
+    if (XEmptyRegion (region))
+    {
+	compLogMessage ("core", CompLogLevelWarn,
+			"Empty box after applying struts, ignoring struts");
+	region->extents = *pBox;
     }
 
     area->x      = region->extents.x1;
@@ -3170,11 +3364,17 @@ updateWorkareaForScreen (CompScreen *s)
     XRectangle workArea;
     BoxRec     box;
     int        i;
+    Bool       workAreaChanged = FALSE;
 
     for (i = 0; i < s->nOutputDev; i++)
-	computeWorkareaForBox (s,
-			       &s->outputDev[i].region.extents,
-			       &s->outputDev[i].workArea);
+    {
+	computeWorkareaForBox (s, &s->outputDev[i].region.extents, &workArea);
+	if (memcmp (&workArea, &s->outputDev[i].workArea, sizeof (XRectangle)))
+	{
+	    workAreaChanged = TRUE;
+	    s->outputDev[i].workArea = workArea;
+	}
+    }
 
     box.x1 = 0;
     box.y1 = 0;
@@ -3185,11 +3385,15 @@ updateWorkareaForScreen (CompScreen *s)
 
     if (memcmp (&workArea, &s->workArea, sizeof (XRectangle)))
     {
-	CompWindow *w;
-
+	workAreaChanged = TRUE;
 	s->workArea = workArea;
 
 	setDesktopHints (s);
+    }
+
+    if (workAreaChanged)
+    {
+	CompWindow *w;
 
 	/* as work area changed, update all maximized windows on this
 	   screen to snap to the new work area */
@@ -3358,9 +3562,10 @@ getActiveWindow (CompDisplay *display,
 				 XA_WINDOW, &actual, &format,
 				 &n, &left, &data);
 
-    if (result == Success && n && data)
+    if (result == Success && data)
     {
-	memcpy (&w, data, sizeof (Window));
+	if (n)
+	    memcpy (&w, data, sizeof (Window));
 	XFree (data);
     }
 
@@ -3676,7 +3881,7 @@ applyStartupProperties (CompScreen *screen,
 
 	workspace = sn_startup_sequence_get_workspace (s->sequence);
 	if (workspace >= 0)
-	    window->desktop = workspace;
+	    setDesktopForWindow (window, workspace);
 
 	window->initialTimestamp    =
 	    sn_startup_sequence_get_timestamp (s->sequence);
@@ -3697,7 +3902,7 @@ sendWindowActivationRequest (CompScreen *s,
     xev.xclient.message_type = s->display->winActiveAtom;
     xev.xclient.window	     = id;
 
-    xev.xclient.data.l[0] = 2;
+    xev.xclient.data.l[0] = ClientTypePager;
     xev.xclient.data.l[1] = 0;
     xev.xclient.data.l[2] = 0;
     xev.xclient.data.l[3] = 0;
@@ -3954,8 +4159,7 @@ clearScreenOutput (CompScreen	*s,
    more than one viewport the most appropriate viewport is returned. How the
    most appropriate viewport is computed can be made optional if necessary. It
    is currently computed as the viewport where the center of the window is
-   located, except for when the window is visible in the current viewport as
-   the current viewport is then always returned. */
+   located. */
 void
 viewportForGeometry (CompScreen *s,
 		     int	x,
@@ -3972,18 +4176,6 @@ viewportForGeometry (CompScreen *s,
     width  += borderWidth * 2;
     height += borderWidth * 2;
 
-    if ((x < s->width  && x + width  > 0) &&
-	(y < s->height && y + height > 0))
-    {
-	if (viewportX)
-	    *viewportX = s->x;
-
-	if (viewportY)
-	    *viewportY = s->y;
-
-	return;
-    }
-
     if (viewportX)
     {
 	centerX = x + (width >> 1);
@@ -4044,17 +4236,26 @@ outputDeviceForGeometry (CompScreen *s,
 
     if (strategy == OUTPUT_OVERLAP_MODE_SMART)
     {
+	int centerX, centerY;
+
 	/* for smart mode, calculate the overlap of the whole rectangle
 	   with the output device rectangle */
 	geomRect.x2 = width + 2 * borderWidth;
 	geomRect.y2 = height + 2 * borderWidth;
 
 	geomRect.x1 = x % s->width;
-	if ((geomRect.x1 + geomRect.x2 / 2) < 0)
+	centerX = geomRect.x1 + (geomRect.x2 / 2);
+	if (centerX < 0)
 	    geomRect.x1 += s->width;
+	else if (centerX > s->width)
+	    geomRect.x1 -= s->width;
+
 	geomRect.y1 = y % s->height;
-	if ((geomRect.y1 + geomRect.y2 / 2) < 0)
+	centerY = geomRect.y1 + (geomRect.y2 / 2);
+	if (centerY < 0)
 	    geomRect.y1 += s->height;
+	else if (centerY > s->height)
+	    geomRect.y1 -= s->height;
 
 	geomRect.x2 += geomRect.x1;
 	geomRect.y2 += geomRect.y1;
@@ -4125,7 +4326,7 @@ outputDeviceForGeometry (CompScreen *s,
 		}
 	    }
     }
-    
+
     return highest;
 }
 
